package ictag;
import gnu.io.CommPortIdentifier;
import gnu.io.SerialPort;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.List;

public class HF04SL {
	private final int RETRY_COUNT = 10;
	private final int rate=38400;
	private final int parity=SerialPort.PARITY_NONE;
	private final int databits=8;
	private final int stopbits=SerialPort.STOPBITS_1;
	private OutputStream output;
	public SerialPort port;

	/**
	 * ICタグ・リーダー／ライターから指定したタグ番号のデータを読み込む。
	 * @param tagNo	タグ番号（０：すべて、1以上で指定されタグ）
	 * @return	ICタグのデータ
	 * @throws IOException	ICタグ接続読込エラー
	 */
	public String[] readICTag(int tagNo) throws IOException {
		StringBuffer buf = new StringBuffer();
		buf.append("2R,");
		buf.append(String.format("%02d,00,0F\r", tagNo));
		write(buf.toString());
		
		String numStr = read(2);
		if (numStr.equals("00")) {
			return (new String[0]);
		}
		else {
			List<String>	list = new ArrayList<String>();
			int numTag = Integer.parseInt(numStr);
			String	str;
			int count = 1;
			for (int i = 0; i < numTag; i++) {
				if (count++ != 1) {
					numStr = read(2);
				}
				str = read(1); // skip ','
				String lenStr = read(2);
				str = read(1); // skip ',' or [CR]
				int len = Integer.parseInt(lenStr, 16);
				if (len > 0) {
					String data = read(len*2);
					str = read(1); // skip ',' or [CR]
					list.add(data);														
				}
			}
			return (list.toArray(new String[0]));
		}		
	}

	/**
	 * 16進数で表現された文字列を文字列に変換する
	 * @param hexStr	16進数文字列
	 * @param index		取得開始位置（0以上のオフセット）
	 * @param len		取得文字列の長さ
	 * @return	変換された文字列
	 */
	public String hexToString(String hexStr, int index, int len) {
		byte[] data = new byte[len];
		for (int i = 0; i < len; i++) {
			data[i] = (byte)Integer.parseInt(hexStr.substring((index+i)*2, (index+i+1)*2), 16);
		}
		return (new String(data));
	}
	
	/**
	 * ICタグをスキャンし、検出されたタグIDの配列を返す
	 * @return	検出されたタグIDの配列
	 * @throws IOException	IOException	ICタグ接続読込エラー
	 */
	public String[] scanICTag() throws IOException {
		write("2XX\r");
		String num = read(2);
		if (num.equals("00"))
			return (null);
		else {
			String[] ids = new String[Integer.parseInt(num)];
			for (int i = 0; i < ids.length; i++) {
				ids[i] = read(16);
				read(1);	// skip ',' or [CR]
			}
			return (ids);
		}
	}
	
	/**
	 * 連続ICタグスキャン（ICカードを検出するまでもどらない）し、タグIDを返す。
	 * @return	タグID
	 * @throws IOException	ICタグ接続読込エラー
	 */
	public String contScanICTag() throws IOException {
		write("2XS\r");
		String res = read(3);
		if (res.equals("01,")) {
			String id = read(16);
			read(1);	// skip [CR]
			return (id);
		}
		else {
			read(2); // skip remain None[CR]
			return (null);
		}
	}
	
	/**
	 * 連続スキャンをキャンセルする
	 * @throws IOException	ICタグ接続読込エラー
	 */
	public void cancelContScanICTag() throws IOException {
		write("a");
	}
	
	/**
	 * ICタグの複数ブロックに連続書き込み。
	 * @param tagNo	タグ番号（1以上）
	 * @param blokNo	ブロック番号
	 * @param data	書き込みデータ
	 * @return	正常に書き込んだか否か
	 * @throws IOException	ICタグ接続書き込みエラー
	 */
	public boolean writeICTag(int tagNo, int blokNo, byte[] data) throws IOException {
		int numBlock = (data.length + 3) / 4;
		StringBuffer buf = new StringBuffer();
		buf.append("2WM,");
		buf.append(String.format("%02x,", tagNo));
		buf.append(String.format("%02x,", blokNo));
		buf.append(String.format("%02x,", numBlock));
		for (int i = 0; i < data.length; i++) 
			buf.append(String.format("%02x", data[i]));
		buf.append("\r");
		write(buf.toString());
		
		read(3); // skip tt,
		String result = read(2);
		read(1); // skip [CR]
		return (result.equals("OK"));
	}
	
	/**
	 * ICタグ・リーダー／ライターとの通信をオープンする
	 * @param pname	シリアルデバイス名
	 * @return	オープンできたか否か
	 */
	public boolean openICTag(String pname){
		if(openSerialPort(pname)){
			String returnCode="";
			for (int i = 0; i < RETRY_COUNT 
				&& (returnCode.indexOf("15693") < 0); i++){
				try{
					write("V\r");
					returnCode = readWithTimer(3000);
				} 
				catch (TimeoutException e1) {}
				catch (Exception e2){
					e2.printStackTrace();
					break;
				}
			}
			return true;
		}
		return false;
	}

	/**
	 * ICタグ・リーダー／ライターとの通信をクローズする
	 */
	public void closeICTag(){
		if(port!=null) port.close();
	}

	/**
	 * ICタグ・リーダー／ライターから読み込むべきデータがあるか否か
	 * @return	データがあるか否か
	 */
	private int available() {
		try {
			InputStream input = port.getInputStream();
			int n = input.available();
			return (n);
		}
		catch (IOException e){
			e.printStackTrace();
		}
		return (0);
	}
	
	/**
	 * 指定バイトの文字列を読み込む。
	 * @param len	読み込む文字列の長さ
	 * @return	読み込んだ文字列
	 */
	public String read(int len) {
		String str = null;
		try {
			InputStream input = port.getInputStream();			
			int remain = len;
			StringBuffer returnBuffer= new StringBuffer();
			while (remain > 0) {
				int n = input.available();
				if (n > remain)	n = remain;
				if (n > 0) {
					byte buf[] = new byte[n];
					input.read(buf);
					returnBuffer.append(new String(buf));
					remain -= n;										
				}
			}
			str = returnBuffer.toString();
		} catch(IOException e){
			e.printStackTrace();
		}
		return str;
	}
	
	/**
	 * タイムアウト指定で文字列を読み込む。
	 * @param timeout	タイムアウト時間（ミリ秒）
	 * @return	読み込んだ文字列
	 * @throws TimeoutException	タイムアウトの場合
	 */
	private String readWithTimer(long timeout) throws TimeoutException{
		try{
			InputStream input = port.getInputStream();
			long start = System.currentTimeMillis();
			int n = input.available();
			StringBuffer returnBuffer= new StringBuffer();
			while(returnBuffer.indexOf("0")==-1){
					long now = System.currentTimeMillis();
					long rest = timeout - (now-start);
					if(rest <= 0){
						throw new TimeoutException("Timeout: " + (now - start));
					}
					try{
						Thread.sleep(10);
					}catch(InterruptedException e){}
				n = input.available();
				byte buf[] = new byte[n];
				input.read(buf);
				returnBuffer.append(new String(buf));
			}
			return returnBuffer.toString();
		}catch(IOException e){
			e.printStackTrace();
		}

		return "";
	}

	/**
	 * シリアルポートをオープンする
	 * @param portName	シリアルポート名
	 * @return	正常にオープンできたか否か
	 */
	private boolean openSerialPort(String portName){
		try{
			CommPortIdentifier portId = CommPortIdentifier.getPortIdentifier(portName);
			port = (SerialPort)portId.open("HF04SLSerialPort", 1000);
		    port.setSerialPortParams(rate, databits, stopbits, parity);
		    output = port.getOutputStream();
		} catch(Exception e){
			return false;
		}
		return true;
	}
	
	/**
	 * シリアルポートへの書き込み
	 * @param str	書き込み文字列
	 * @throws IOException	書き込みエラー
	 */
	private void write(String str) throws IOException{
		write(str.getBytes());
	}

	/**
	 * シリアルポートへの書き込み。
	 * @param bytes	書き込みデータ
	 * @throws IOException	書き込みエラー
	 */
	private void write(byte bytes[]) throws IOException{
		output.write(bytes);
		output.flush();
	}
	
	public static void main(String[] args) {
		HF04SL ictag = new HF04SL();
		try {
			ictag.openICTag("/dev/cu.usbserial-A5002yHm");
			System.out.println(ictag.contScanICTag());

			System.out.println(ictag.writeICTag(1, 1, "0123456789ABCDEF".getBytes()));
			String[] tagData = ictag.readICTag(1);
			System.out.println(tagData[0]);	
			System.out.println(ictag.hexToString(tagData[0], 4, 16));
		}
		catch (Exception e) {	
			e.printStackTrace();
		}
		finally {
			ictag.closeICTag();
		}
	}
}
